//---------------------------------------------------------------------------
// Copyright 2004 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/libAHmath/ReduceSpectrum.c,v $
// $Id: ReduceSpectrum.c,v 1.1 2005/02/17 22:24:32 ah Exp $ 
//
// Description: create frequency spectrum of a wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: ReduceSpectrum.c,v 1.1 2005/02/17 22:24:32 ah Exp $ (c) 2004 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
#include <AH/Math/ReduceSpectrum.h> // own software interface
USING_NAMESPACE(AH)

#include <iostream>
#include <cstdlib>
#include <cmath>
#include <drfftw.h> 
USING_NAMESPACE(std)

#include <AH/IO/Smart/Buffer.h>
USING_NAMESPACE(AH)

BEGIN_NAMESPACE(AH)

// --- local definitions ---

// --- local constants ---

// --- local variables ---

// --- local classes ---

// --- local functions ---

// --- member functions ---

// constructor
ReduceSpectrum::ReduceSpectrum(int v, unsigned int osize, double fmax, 
			      double flo, double fhi) 
    : verbose(v), outsize(osize), fMax(fmax),
      f1(flo), f2(fhi)
{
    if (!this->outsize)
	throw ErrorInternBadArg("ReduceSpectrum constructor: outsize");
    if (this->f1 < 0)
	throw ErrorInternBadArg("ReduceSpectrum constructor: flo");
    if (this->f2 <= 0)
	throw ErrorInternBadArg("ReduceSpectrum constructor: fhi");
    if (this->fMax <= 0)
	throw ErrorInternBadArg("ReduceSpectrum constructor: fmax");
}

// keep linear frequency scale, reduce output size.
void ReduceSpectrum::processLinXScale(const double* data, unsigned int insize,
				      double* avout, double* pkout)
{
    if (!avout && !pkout) return;
    if (!insize)
	throw ErrorInternBadArg("ReduceSpectrum::processLinXScale: isize");

    if (this->verbose > 1) cerr
	<< "ReduceSpectrum::processLinXScale: make lin scale for frequency" 
	<< endl;

    const double fin_inc = this->fMax/(double)(insize - 1);
    double fout = this->f1;
    const double fout_inc = (this->f2 - this->f1) / (this->outsize - 1.0);
    const double flim_inc = 0.5 * fout_inc;

    for (unsigned int out = 0; out < this->outsize; ++out, fout += fout_inc)
    {
	// determine index span (float) of this output frequency:
	double fi1 = (fout - flim_inc) / fin_inc;
	double fi2 = (fout + flim_inc) / fin_inc;

	// determine first and last index:
	int bi1 = (int)(fi1 + 0.5);
	int bi2 = (int)(fi2 + 0.5);

	// determine proportional parts of first and last index:
	double pi1; // proportional part of bi1
	double pi2; // proportional part of bi2
	if (bi1 == bi2)
	{
	    pi1 = 0.5 * fabs(fi2 - fi1);
	    pi2 = pi1;
	}
	else
	{
	    pi1 = (double)bi1 + 0.5 - fi1;
	    pi2 = fabs(fi2 - (double)bi2 + 0.5);
	}

	// sum up over all indices:
	double max = data[bi1] > data[bi2] ? data[bi1] : data[bi2];
	double sum = pi1 * data[bi1] + pi2 * data[bi2];
	double cnt = pi1 + pi2;
	for (int i = bi1+1; i < bi2; ++i)
	{
	    if (data[i] > max) max = data[i];
	    sum += data[i];
	    ++cnt;
	}
	if (avout) avout[out] = sum/cnt;
	if (pkout) pkout[out] = max;
    }

} // end of ReduceSpectrum::processLinXScale()



// convert linear frequency scale into log scale, thereby reducing
// output size.
void ReduceSpectrum::processLogXScale(const double* data,unsigned int insize,
				      double* avout, double* pkout,
				      double smooth)
{
    if (!avout && !pkout) return;
    if (!insize)
	throw ErrorInternBadArg("ReduceSpectrum::processLogXScale: isize");
    if (this->f1 <= 0)
	throw ErrorInternBadArg("ReduceSpectrum::processLogXScale: flo");

    if (this->verbose > 1) cerr
	<< "ReduceSpectrum::processLogXScale: make log scale for frequency" 
	<< endl;

    const double fin_inc = this->fMax/(double)(insize - 1);
    double fout = this->f1;
    const double fout_inc = pow(this->f2/this->f1, 1.0 /
				(this->outsize - 1.0));
    const double flim_inc = pow(this->f2/this->f1, 0.5 /
				(this->outsize - 1.0));

    for (unsigned int out = 0; out < this->outsize; ++out, fout *= fout_inc)
    {
	// determine frequency span of this output frequency:
	double fs1 = fout / flim_inc;
	double fs2 = fout * flim_inc;
	if (fs1 < 0) fs1 = 0;
	if (fs2 > this->fMax) fs2 = this->fMax;

	// determine index span (float) of this output frequency:
	double fi1 = fs1 / fin_inc;
	double fi2 = fs2 / fin_inc;

	// determine first and last index:
	int bi1 = (int)(fi1 + 0.5);
	int bi2 = (int)(fi2 + 0.5);
	if (bi1 < 0) bi1 = 0;
	if (bi2 >= (int)insize) bi1 = insize - 1;

	// determine proportional parts of first and last index:
	double pi1; // proportional part of bi1
	double pi2; // proportional part of bi2
	if (bi1 == bi2)
	{
	    pi1 = 0.5 * fabs(fi2 - fi1);
	    pi2 = pi1;
	}
	else
	{
	    pi1 = (double)bi1 + 0.5 - fi1;
	    pi2 = fabs(fi2 - (double)bi2 + 0.5);
	}

	// sum up over all indices:
	double max = data[bi1] > data[bi2] ? data[bi1] : data[bi2];
	double sum = pi1 * data[bi1] + pi2 * data[bi2];
	double cnt = pi1 + pi2;
	for (int i = bi1+1; i < bi2; ++i)
	{
	    sum += data[i];
	    if (data[i] > max) max = data[i];
	    ++cnt;
	}
	// I do not really understand why we must divide the sum
	// by sqrt(cnt), but if we don't, then a pink noise spectrum
	// rises by 10 dB per octave. 
	if (avout) avout[out] = sum / sqrt(cnt);
	if (pkout) pkout[out] = max;
    }

    if (avout && (smooth > 0))
	this->smoothLogXScale(avout, smooth);

} // end of ReduceSpectrum::processLogXScale()



// smooth data in logarithmic frequency scale
void ReduceSpectrum::smoothLogXScale(double* data, double smooth)
{
    if (this->verbose > 1) cerr
	<< "ReduceSpectrum::smoothLogXScale: smooth log scale for frequency" 
	<< endl;

    IOSmartBuffer<double> dout(this->outsize);

    // convert number of octaves into range factor:
    // (a) fs2 = 2^smooth * fs1
    // (b) sqrt(fs1 * fs2) = f
    // ==> fs1 = f / sqrt(2^smooth) and fs2 = f * sqrt(2^smooth)
    double range = sqrt(pow(2, smooth));

    double f = this->f1;
    const double f_inc = pow(this->f2/this->f1, 1.0 / (this->outsize - 1.0));
    const double log_f_inc = log(f_inc);
    for (unsigned int out = 0; out < this->outsize; ++out, f *= f_inc)
    {
	// determine frequency span of smoothing range:
	double fs1 = f / range;
	double fs2 = f * range;
	if (fs1 < f1) fs1 = this->f1;
	if (fs2 > f2) fs2 = this->f2;

	// determine index span (float) of smoothing range:
	double fi1 = log(fs1/this->f1) / log_f_inc;
	double fi2 = log(fs2/this->f1) / log_f_inc;

	// determine first and last index:
	int bi1 = (int)(fi1 + 0.5);
	int bi2 = (int)(fi2 + 0.5);
	if (bi1 < 0) bi1 = 0;
	if (bi2 >= (int)this->outsize) bi1 = this->outsize - 1;

	// determine proportional parts of first and last index:
	double pi1; // proportional part of bi1
	double pi2; // proportional part of bi2
	if (bi1 == bi2)
	{
	    pi1 = 0.5 * fabs(fi2 - fi1);
	    pi2 = pi1;
	}
	else
	{
	    pi1 = (double)bi1 + 0.5 - fi1;
	    pi2 = fabs(fi2 - (double)bi2 + 0.5);
	}

	// sum up over all indices:
	double sum = pi1 * data[bi1] + pi2 * data[bi2];
	double cnt = pi1 + pi2;
	for (int i = bi1+1; i < bi2; ++i)
	{
	    sum += data[i];
	    ++cnt;
	}
	dout[out] = sum / cnt;
    }

    // copy result back into input buffer, update input buffer size
    for (unsigned int i = 0; i < this->outsize; ++i)
    {
	data[i] = dout[i];
    }

} // end of ReduceSpectrum::smoothLogXScale()


END_NAMESPACE

// --- end of file ---
